from pathlib import Path

from lod.client import LocustClient
from lod.manager import LocustBaseManager, LocustSingleNodeManager, LocustDistributedManager
from lod.proxy import get_proxy_settings_for_port


class LocustRunner:

    def __init__(self, locustfile_path: str | Path, port: int = 8089):
        self._web_port = port
        self._locustfile_path = locustfile_path
        self._distributed = False
        self._locust_manager: LocustBaseManager = LocustSingleNodeManager(
            file_name=locustfile_path, web_port=port
        )
        self._is_locust_running = False
        self._locust_client = LocustClient(
            server_port=port,
        )
        self._proxy_settings = get_proxy_settings_for_port(self._web_port)
        self._preloaded_locust_swarm = None

    def start_locust(self) -> int:
        pid = self._locust_manager.start()
        self._is_locust_running = True
        try:
            import IPython
            display_html = IPython.get_ipython().user_ns["displayHTML"]
            display_text = f'<a href="{self._proxy_settings.get_proxy_url(ensure_ends_with_slash=True)}">Click to go to Access Locust Web UI!</a>'
            display_html(display_text)
        except Exception:
            print(f"Access Locust Web UI at {self._proxy_settings.get_proxy_url(ensure_ends_with_slash=True)}")
        if self._preloaded_locust_swarm:
            print("Preloaded swarm parameters:", self._preloaded_locust_swarm)
            self.run_swarm(**self._preloaded_locust_swarm)
            self._preloaded_locust_swarm = None
        return pid

    def distributed(self, process_to_core_count_ratio: float = 2.0):
        self._distributed = True
        self._locust_manager = LocustDistributedManager(
            file_name=self._locustfile_path,
            web_port=self._web_port,
            process_to_core_count_ratio=process_to_core_count_ratio
        )
        return self

    def stop_locust(self):
        self._locust_manager.kill()
        self._is_locust_running = False

    def set_initial_swarm(self,
              host: str,
              user_count: int,
              spawn_rate: int,
              run_time: str = "5m"):
        if self._is_locust_running:
            raise Exception("Locust is already running. Please stop it before setting initial swarm parameters.")
        else:
            self._preloaded_locust_swarm = {
                'host': host,
                'user_count': user_count,
                'spawn_rate': spawn_rate,
                'run_time': run_time
            }

        return self

    def run_swarm(self,
                  host: str,
                  user_count: int,
                  spawn_rate: int,
                  run_time: str = "5m"):
        if self._is_locust_running:
            self._locust_client.swarm(
                host=host,
                user_count=user_count,
                spawn_rate=spawn_rate,
                run_time=run_time
            )
        else:
            raise Exception("Locust is not running. Please start Locust first.")

    def stop_swarm(self):
        if self._is_locust_running:
            self._locust_client.stop_swarm()
        else:
            raise Exception("Locust is not running. Please start Locust first.")


import abc
import time

import psutil
import socket
import ipaddress
import os
import signal
import subprocess

import requests
from requests import RequestException


def get_rfc_1918_network_ip():
    for interface, snics in psutil.net_if_addrs().items():
        for snic in snics:
            if snic.family == socket.AF_INET:
                ip = snic.address
                # Skip loopback addresses
                if ip.startswith("127."):
                    continue
                try:
                    # Validate if the IP is private (RFC1918)
                    if ipaddress.ip_address(ip).is_private:
                        return ip
                except Exception:
                    continue
    # Fallback in case no suitable IP is found
    return None


def is_process_running(host: str, port: int, timeout: int = 10) -> bool:
    start_time = time.time()
    url = f"http://{host}:{port}/logs"

    while time.time() - start_time < timeout:
        try:
            response = requests.get(url, timeout=1)
            if response.status_code == 200:
                print("Service is up and running.")
                return True
        except RequestException:
            pass  # Connection failed, will retry

        print("Service not yet running, retrying in 1 second...")
        time.sleep(1)  # Wait 1 second before retrying

    return False  # Timed out without getting a 200 response


def is_cluster_running(host: str, port: int, timeout: int = 10) -> bool:
    start_time = time.time()
    url = f"http://{host}:{port}/logs"

    while time.time() - start_time < timeout:
        try:
            response = requests.get(url, timeout=1)
            if response.status_code == 200:
                resp_dict = response.json()
                workers = resp_dict.get("workers")
                if workers and len(workers) > 0:
                    print("Cluster is up and running.")
                    return True
        except RequestException:
            pass  # Connection failed, will retry

        print("Cluster not yet ready, retrying in 1 second...")
        time.sleep(1)  # Wait 1 second before retrying

    return False  # Timed out without getting a 200 response


class LocustUtils:

    @staticmethod
    def is_running_on_current_node() -> bool:
        for proc in psutil.process_iter(['cmdline']):
            try:
                cmdline = proc.info['cmdline']
                if cmdline and any("locust" in part.lower() for part in cmdline):
                    return True
            except (psutil.NoSuchProcess, psutil.AccessDenied):
                continue
        return False

    @staticmethod
    def kill_on_current_node() -> bool:
        killed = False
        for proc in psutil.process_iter(['pid', 'cmdline']):
            try:
                cmdline = proc.info['cmdline']
                if cmdline and any("locust" in part.lower() for part in cmdline):
                    os.kill(proc.pid, signal.SIGTERM)
                    killed = True
            except (psutil.NoSuchProcess, psutil.AccessDenied):
                continue
        return killed

    @staticmethod
    def start_standard_on_current_node(file_name: str = "locustfile.py", web_port: int = 8089) -> int:
        master_cmd = f"locust -f {file_name} --web-port {web_port}"
        process = subprocess.Popen(master_cmd.split())
        print(f"Locust started with pid: {process.pid}")
        # Wait for the process to start
        is_process_running("0.0.0.0", web_port)
        return process.pid

    @staticmethod
    def start_driver_on_current_node(file_name: str = "locustfile.py", web_port: int = 8089) -> int:
        master_cmd = f"locust -f {file_name} --master --web-port {web_port}"
        process = subprocess.Popen(master_cmd.split())
        # Wait for the process to start
        is_process_running("0.0.0.0", web_port)
        return process.pid

    @staticmethod
    def start_worker_on_current_node(driver_ip: str, process_to_core_count_ratio: float = 2.0) -> int:
        import multiprocessing
        num_processes = int(multiprocessing.cpu_count() * process_to_core_count_ratio)
        worker_cmd = f"locust -f - --worker --master-host {driver_ip} --processes {num_processes}"
        process = subprocess.Popen(worker_cmd.split())
        return process.pid


class LocustBaseManager(abc.ABC):

    @abc.abstractmethod
    def is_running(self) -> bool:
        pass

    @abc.abstractmethod
    def kill(self) -> bool:
        pass

    @abc.abstractmethod
    def start(self) -> int:
        pass


class LocustSingleNodeManager(LocustBaseManager):

    def __init__(self, file_name: str = "locustfile.py", web_port: int = 8089):
        self._file_name = file_name
        self._web_port = web_port

    def is_running(self) -> bool:
        return LocustUtils.is_running_on_current_node()

    def kill(self) -> bool:
        return LocustUtils.kill_on_current_node()

    def start(self) -> int:
        return LocustUtils.start_standard_on_current_node(
            file_name=self._file_name, web_port=self._web_port
        )


def get_spark():
    from pyspark.sql import SparkSession
    return SparkSession.getActiveSession()


def active_worker_count():
    spark = get_spark()
    return int(spark.conf.get("spark.databricks.clusterUsageTags.clusterWorkers", "1"))


class LocustDistributedManager(LocustBaseManager):

    def __init__(self,
                 file_name: str = "locustfile.py",
                 web_port: int = 8089,
                 process_to_core_count_ratio: float = 2.0):
        self._file_name = file_name
        self._web_port = web_port
        self._process_to_core_count_ratio = process_to_core_count_ratio
        self._active_number_of_workers = active_worker_count()
        self._worker_names = [f"worker-{i}" for i in range(self._active_number_of_workers)]

    def is_running(self) -> bool:
        running_states = {"driver": LocustUtils.is_running_on_current_node()}

        def is_locust_running_on_executor(_executor: str):
            ip = get_rfc_1918_network_ip()
            status = LocustUtils.is_running_on_current_node()
            return _executor, ip, status

        results = get_spark() \
            .sparkContext \
            .parallelize(self._worker_names, self._active_number_of_workers) \
            .map(is_locust_running_on_executor).collect()

        for executor, ip, status in results:
            running_states[f"{executor}-{ip}"] = status

        for key, value in running_states.items():
            print(f"Locust running on {key}: {value}")

        return all(running_states.values())

    def kill(self) -> bool:
        killed_states = {"driver": LocustUtils.kill_on_current_node()}

        def kill_on_executor(_executor: str):
            ip = get_rfc_1918_network_ip()
            status = LocustUtils.kill_on_current_node()
            return _executor, ip, status

        results = get_spark() \
            .sparkContext \
            .parallelize(self._worker_names, self._active_number_of_workers) \
            .map(kill_on_executor).collect()

        for executor, ip, status in results:
            killed_states[f"{executor}-{ip}"] = status

        for key, value in killed_states.items():
            print(f"Locust killed on {key}: {value}")

        return all(killed_states.values())

    def start(self) -> int:
        driver_ip = get_rfc_1918_network_ip()

        driver_pid = LocustUtils.start_driver_on_current_node(file_name=self._file_name, web_port=self._web_port)
        print(f"Driver, IP: {driver_ip} - Locust master started with pid: {driver_pid}.")

        def start_locust_worker(_executor_name: str, _driver_ip: str, _process_to_core_count_ratio: float = 2.0):
            worker_ip = get_rfc_1918_network_ip()
            worker_pid = LocustUtils.start_worker_on_current_node(_driver_ip, _process_to_core_count_ratio)
            return _executor_name, worker_ip, worker_pid

        # Distribute the worker start command using map to collect executor info
        results = get_spark() \
            .sparkContext \
            .parallelize(self._worker_names, self._active_number_of_workers) \
            .map(lambda executor_name: start_locust_worker(executor_name, driver_ip, self._process_to_core_count_ratio)) \
            .collect()

        for executor, ip, pid in results:
            print(f"Executor {executor}, IP: {ip} - Locust worker started with PID: {pid}")

        is_cluster_running("0.0.0.0", self._web_port, timeout=10)

        return driver_pid



from dataclasses import dataclass
from typing import Optional
from urllib.parse import urlparse


def remove_lowest_subdomain_from_host(url):
    parsed_url = urlparse(url)
    host = parsed_url.netloc if parsed_url.netloc else parsed_url.path
    parts = host.split('.')
    # Check if there are subdomains to remove
    if len(parts) > 2:
        # Remove the lowest subdomain
        parts.pop(0)

    # Reconstruct the modified host
    modified_host = '.'.join(parts)

    return modified_host


class ClusterUtils:

    def __init__(self):
        from pyspark.sql import SparkSession
        self._spark = SparkSession.getActiveSession()

    @property
    def workspace_url(self):
        return self._spark.conf.get("spark.databricks.workspaceUrl")

    @property
    def org_id(self):
        return self._spark.conf.get("spark.databricks.clusterUsageTags.orgId")

    @property
    def cluster_id(self):
        return self._spark.conf.get("spark.databricks.clusterUsageTags.clusterId")

    @property
    def cloud(self):
        url = self.workspace_url
        if url.endswith("azuredatabricks.net"):
            return "azure"
        if ".gcp." in url:
            return "gcp"
        return "aws"


@dataclass
class ProxySettings:
    proxy_url: str
    port: str
    url_base_path: str
    url_base_path_no_port: Optional[str] = None

    def get_proxy_url(self, ensure_ends_with_slash=False):
        """
        For certain apps that use relative paths like "assets/index-*.js" we need to ensure that the url ends
        with a slash.
        """
        if ensure_ends_with_slash is True:
            return self.proxy_url.rstrip("/") + "/"
        return self.proxy_url


def get_cloud_proxy_settings(cloud: str,
                             org_id: str,
                             cluster_id: str,
                             port: int,
                             host: str) -> ProxySettings:
    cloud_norm = cloud.lower()
    if cloud_norm not in ["aws", "azure"]:
        raise Exception("only supported in aws or azure")
    prefix_url_settings = {
        "aws": "https://dbc-dp-",
        "azure": "https://adb-dp-",
    }
    suffix_url_settings = {
        "azure": "azuredatabricks.net",
    }
    if cloud_norm == "aws":
        suffix = remove_lowest_subdomain_from_host(host)
        suffix_url_settings["aws"] = suffix

    org_shard = ""
    # org_shard doesnt need a suffix of "." for dnsname its handled in building the url
    # only azure right now does dns sharding
    # gcp will need this
    if cloud_norm == "azure":
        org_shard_id = int(org_id) % 20
        org_shard = f".{org_shard_id}"

    url_base_path_no_port = f"/driver-proxy/o/{org_id}/{cluster_id}"
    url_base_path = f"{url_base_path_no_port}/{port}/"
    return ProxySettings(
        proxy_url=f"{prefix_url_settings[cloud_norm]}{org_id}{org_shard}.{suffix_url_settings[cloud_norm]}{url_base_path}",
        port=str(port),
        url_base_path=url_base_path,
        url_base_path_no_port=url_base_path_no_port
    )


def get_proxy_settings_for_port(port: int) -> ProxySettings:
    utils = ClusterUtils()
    print("Proxy Inputs", utils.cloud, utils.cluster_id, utils.org_id, utils.workspace_url)
    return get_cloud_proxy_settings(utils.cloud, utils.org_id, utils.cluster_id, port, utils.workspace_url)



# Databricks notebook source
# will publish to pypi soon
%pip install git+https://github.com/stikkireddy/locust-on-databricks.git --upgrade
%restart_python

# COMMAND ----------

# locust-on-databricks aka lod
from lod import LocustRunner

# COMMAND ----------

runner = LocustRunner(locustfile_path="locustfile.py")

runner.distributed().set_initial_swarm(
  host="https://google.com",
  user_count=30,
  spawn_rate=2,
  # run_time = "5m"
).start_locust()

# COMMAND ----------

# Kill an Existing Swarm
# runner.stop_swarm()

# COMMAND ----------

runner.stop_locust()

# COMMAND ----------

# Optionally Run your Own Swarm
# runner.run_swarm(
#   host="https://google.com",
#   user_count=10,
#   spawn_rate=2,
#   # run_time = "5m"
# )


from setuptools import setup, find_packages

with open("README.md", "r", encoding="utf-8") as fh:
    long_description = fh.read()

with open("requirements.txt", "r", encoding="utf-8") as fh:
    install_requires = fh.read().splitlines()


setup(
    name="locust-on-databricks",
    author="Sri Tikkireddy",
    author_email="sri.tikkireddy@databricks.com",
    description="Run locust on single node or distributed on Databricks.",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/stikkireddy/locust-on-databricks",
    packages=find_packages(),
    install_requires=install_requires,
    extras_require={
            'dev': [
                'pytest',
                'pytest-pep8',
                'pytest-cov',
                'pyspark==3.4.1',
                'twine'
            ]
        },
    setup_requires=["setuptools_scm"],
    use_scm_version=True,
    classifiers=[
        "Programming Language :: Python :: 3",
        "Operating System :: OS Independent",
    ],
    python_requires=">=3.11",
)



import requests

class LocustClient:

    def __init__(self, server_host: str = "0.0.0.0", server_port: int = 8089):
        self._session = requests.Session()
        self._server_host = server_host
        self._server_port = server_port
        self._base_url = f"http://{server_host}:{server_port}"

    def swarm(self,
              host: str,
              user_count: int,
              spawn_rate: int,
              run_time: str = "5m") -> requests.Response:
        """
        Start a Locust swarm against the specified host.

        Args:
            host: Target URL to test against (required)
            user_count: Number of simulated users
            spawn_rate: Rate at which users spawn per second
            run_time: How long the test should run (e.g. '5m', '30s')

        Returns:
            Response from the Locust server
        """
        endpoint = f"{self._base_url}/swarm"

        data = {
            'user_count': user_count,
            'spawn_rate': spawn_rate,
            'host': host,
            'run_time': run_time
        }

        try:
            response = self._session.post(endpoint, data=data)
            response.raise_for_status()  # Raise exception for 4XX/5XX responses
            print(f"Swarm started successfully: {response.text}")
            return response
        except requests.RequestException as e:
            print(f"Error starting swarm: {str(e)}")
            raise

    def stop_swarm(self) -> requests.Response:
        """
        Stop the Locust swarm.

        Returns:
            Response from the Locust server
        """
        endpoint = f"{self._base_url}/stop"

        try:
            response = self._session.get(endpoint)
            response.raise_for_status()  # Raise exception for 4XX/5XX responses
            print(f"Swarm stopped successfully: {response.text}")
            return response
        except requests.RequestException as e:
            print(f"Error stopping swarm: {str(e)}")
            raise